// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      GtiCorr.cc: --- GTI correction for epic_lccorr
//
//      DESCRIPTION:
//             Corrects the light curve time bin from GTIs
//
//      CHANGE HISTORY:
//              AI, 2008-01-11: 1.0.0: - first version
//
//      AUTHOR:
//              A. Ibarra (AI), Wed Jan  9 17:54:04 GMT 2008
//
//

#include <vector>
#include <iomanip>
#include "GtiCorr.h"


GtiCorr::GtiCorr(GtiList gti) : _gtiList(gti)
{

}

void GtiCorr::doCorr(LCGrid *lcgrid)
{
  std::vector<double> totalGtiBin(int(lcgrid->length()),0.0);
  unsigned int n = 0; // Gti counting number

  GtiList *gtiList = new GtiList();

  if (_gtiList.size() == 0)
	  errstr << "GTI extension empty" << Sas::error("EmptyGTI");

  // Loop over all the GTIs
  while (n < _gtiList.size() )
  {
	  double tStart = 0.0; // Valid time start
	  double tStop = 0.0;  // Valid time stop

	  // LC bin that correspond to the start time of the current GTI
	  int currentBin = lcgrid->cell(_gtiList[n].start());
	  // LC bin that correspond to the stop time of the current GTI
	  int binStop  = lcgrid->cell(_gtiList[n].stop());

	  // Check if current gti falls partly outside the first lc bin
	  if ( (currentBin < 0) && (binStop >= 0))
	  {
		  currentBin = 0;
		  tStart = lcgrid->cellLimits(currentBin)[0];
	  }

	  // Counter to know the number of LC bin covered by the GTI
	  // binCounter greater than 1 means that the GTI extends over more that 1 LC bin
	  int binCounter = binStop - currentBin;

	  errstr << std::setprecision(14) << "gti["<<n<<"]= ( "<<_gtiList[n].start()<<"," <<_gtiList[n].stop()<<" ) "
	  	<< Sas::message(Msg::Lib, Msg::Verbose);

	  // First condition: loop over all possible LC bin. If it is 0 means that the GTI falls into
	  //                  a single LC bin
	  // Second condition: Check if GTI happens before the LC start time
	  // Third condition:  Check  if GTI starts before the LC stop time
	  // Fourth condition: Check if the latest GTI ends after the LC stop time
	  while ( (binCounter >= 0) && (currentBin >= 0) &&
			  ( _gtiList[n].start() < lcgrid->_stop) && (tStart != lcgrid->_stop) )
	  {
		  vector<double> currentLimits = lcgrid->cellLimits(currentBin);

		  errstr << std::setprecision(14) << "LC bin limits= ( "<< currentLimits[0]<<","<<currentLimits[1]<<" )"
		  	<< Sas::message(Msg::Lib, Msg::Verbose);

		  if ( _gtiList[n].start() >= currentLimits[0] )
			  tStart = _gtiList[n].start();

		  if ( _gtiList[n].stop() > currentLimits[1] )
			  tStop = currentLimits[1];
		  else
			  tStop = _gtiList[n].stop();

		  Gti gti(tStart, tStop);
		  gtiList->push_back(gti);

		  totalGtiBin[currentBin] += tStop - tStart;
		  errstr << "totalGtiBin["<<currentBin<<"]= "<< totalGtiBin[currentBin]
		    << Sas::message(Msg::Lib, Msg::Verbose);

		  // Move the current start time to the beginning of the next LC bin
		  tStart = currentLimits[1];

		  // Move to the next LC bin
		  currentBin++;

		  binCounter--;
	  }
	  n++;
  } // end while GTIs

  // Override the existing default time serie GTI list with the new one.
  lcgrid->setGtiList(gtiList);

  // Calculate the corresponding fraction.
  for(unsigned int i = 0 ; i < lcgrid->length(); i++)
  {
	  _gtiCorr.push_back(totalGtiBin[i]/lcgrid->getBinSize(i));
	  lcgrid->_fracExp[i] = (totalGtiBin[i]/lcgrid->getBinSize(i));
	  (*(lcgrid->_effBinTime))[i] = totalGtiBin[i];
  }


}

void GtiCorr::printCorr()
{
	for (unsigned int i = 0 ; i < _gtiCorr.size() ; i++ )
	{
		errstr <<  "GTICorr["<<i<<"]= "<< _gtiCorr[i]
		         << Sas::message(Msg::Lib, Msg::Sparse);
	}
}

