// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      LCGrid.cc: --- Class related to the time serie binning
//
//      DESCRIPTION:
//             Interface to the common functionalities of a binning class.
//
//      CHANGE HISTORY:
//              AI, 2008-01-11: 1.0.0: - first version
//
//      AUTHOR:
//              A. Ibarra (AI), Wed Jan  9 17:54:04 GMT 2008
//
//
#include <iomanip>
#include <iostream>
#include "LCGrid.h"
#include "errstr.h"

LCGrid::LCGrid(double val0, double val1, size_t cells)
{
  _stride = (val1 - val0) / cells;
  _origin = val0 + _stride / 2.0;
  _cells = cells - 1; //First bin is numbered as 0!!!
  _min = 0;
  _max = cells;

  _start = val0;
  _stop = val1;

  /*
   * Create the original GTI list. Before any correction
   * it is the orginal light curve time bin.
   */
  double stop = 0.0;
  double start = _start;
  _gtiList = new GtiList();
  for (unsigned int i = 1 ;  i <= _cells ; i++)
  {
	  stop = _start + i*_stride;
	  Gti gti(start,stop);
	  _gtiList->push_back(gti);
	  start = stop;
  }
  Gti gti(start,_stop);
  _gtiList->push_back(gti);


  // Initialize _fracExpExp vector
  _effBinTime = new std::vector<double>;
  for (unsigned int i = 0 ; i<=_cells ; i++)
  {
	  _fracExp.push_back(0.0);
	  _fracExpNoArea.push_back(0.0);
	  _effBinTime->push_back(0.0);
  }
}

LCGrid::LCGrid(double val0, double val1, double binSize)
{
  _stride = binSize;
  _origin = val0 + _stride / 2.0;
  _min = 0;
  _cells = size_t(ceil((val1 - val0) / _stride)) - 1;
  _max = _cells ;

  _start = val0;
  _stop =  val1;
  /*
    * Create the original GTI list. Before any correction
    * it is the orginal light curve time bin.
    */
   double stop = 0.0;
   double start = _start;
   _gtiList = new GtiList();
   for (unsigned int i = 1 ;  i <= _cells ; i++)
   {
	   stop = _start + i*_stride;
 	  Gti gti(start,stop);
 	  _gtiList->push_back(gti);
 	  start = stop;
   }
   Gti gti(start,_stop);
   _gtiList->push_back(gti);

   // Initialize _fracExp vector
   _effBinTime = new std::vector<double>;
    for (unsigned int i = 0 ; i<=_cells ; i++)
    {
      _fracExp.push_back(0.0);
      _fracExpNoArea.push_back(0.0);
      _effBinTime->push_back(0.0);
    }

}

LCGrid::~LCGrid()
{
	delete _gtiList;
	delete _effBinTime;
	errstr  << "Cleaning Light Curve vectors"<< Sas::message(Msg::Lib, Msg::Verbose);

	//_gtiList.clear();
	_fracExp.clear();
	_fracExpNoArea.clear();
	//_effBinTime.clear();
	_counts.clear();

}

double LCGrid::minval() const
{
  return _max < _min ? _origin : (*this)[_stride < 0 ? _max + .5 : _min - 0.5] ;
}

double LCGrid::maxval() const
{
  return _max < _min ? _origin : (*this)[_stride < 0 ? _min - 0.5 : _max + 0.5];
}

double LCGrid::size(size_t n) const
{
  double tick = (*this)[n];

  double start = tick - _stride / 2.0;

  double stop = tick + _stride / 2.0;

  if (stop > _stop)
    stop = _stop;

  return stop - start;
}

std::vector<double> LCGrid::cellLimits(size_t n) const
{
  double tick = (*this)[n];

  double start = tick - _stride / 2.0;

  double stop = tick + _stride / 2.0;

  if (stop > _stop)
    stop = _stop;

  std::vector<double> limits;
  limits.push_back(start);
  limits.push_back(stop);

  return limits;
}

double LCGrid::getBinSize(size_t n) const
{
  std::vector<double> limits = this->cellLimits(n);

  return limits[1]-limits[0];
}

void LCGrid::setFracExp(std::vector<double> fracExp)
{
	this->_fracExp.clear();
	this->_fracExp = fracExp;

}



void LCGrid::printFracExp()
{
  for (unsigned int i = 0 ; i < this->_fracExp.size() ; i++)
    cout << setprecision(10)<< "FracExp["<<i<<"]= "<< this->_fracExp[i] << endl;

}

void LCGrid::setGtiList(GtiList* gtiList)
{
	delete _gtiList;

	_gtiList = gtiList;
}

GtiList* LCGrid::getGtiList()
{
	return _gtiList;
}

void LCGrid::printGtiList()
{
	  for (unsigned int i = 0 ; i < _gtiList->size() ; i++)
	 	  cout <<setprecision(18)<< (*_gtiList)[i].start() << " " << (*_gtiList)[i].stop() << endl;
}


void LCGrid::setCounts(int32 *counts)
{
	for (unsigned int i = 0 ; i <= this->_cells ; i++)
	{
		 this->_counts.push_back((double) counts[i]) ;
	}

}

void LCGrid::setCounts(real64 *rate)
{
	for (unsigned int i = 0 ; i <= this->_cells ; i++)
	{
	  this->_counts.push_back( rate[i]*this->_stride );
	}

}

std::vector<double> LCGrid::getCounts()
{
	return this->_counts;
}

void LCGrid::printCounts()
{
	for (unsigned int i = 0 ; i <= this->_cells ; i++)
	{
		cout << "val " << this->_counts[i] << endl;
	}

}

void LCGrid::printLC()
{
	double rate, error;
	for (unsigned int i = 0 ; i <= this->_cells ; i++)
	{
		rate = this->_fracExp[i] == 0.0 ? 0.0 : this->_counts[i]/(this->_stride *this->_fracExp[i]);
		error = this->_fracExp[i] == 0.0 ? 0.0 : sqrt(this->_counts[i])/(this->_stride *this->_fracExp[i]);
		_rateCorr.push_back(rate);
		_errRateCorr.push_back(error);
		_timeStamp.push_back((*this)[i]);
		cout << setprecision(19)<< (*this)[i] <<" " << rate<< " " << error<<endl;
	}
}

void LCGrid::calRateAndErr()
{
	_rateCorr.clear();
	_errRateCorr.clear();
	_timeStamp.clear();
	double rate, error;

	for (unsigned int i = 0 ; i <= this->_cells ; i++)
	{
	  rate = this->_fracExp[i] == 0.0 ? 0.0 : this->_counts[i]/(this->_stride *this->_fracExp[i]);
	  error = this->_fracExp[i] == 0.0 ? 0.0 : sqrt(std::abs(this->_counts[i]))/(this->_stride *this->_fracExp[i]);
	  _rateCorr.push_back(rate);
	  _errRateCorr.push_back(error);
	  _timeStamp.push_back((*this)[i]);
	}

}

void LCGrid::calFracExpWeight()
{
  std::vector<double> res;
  for (unsigned int i = 0 ; i < this->_counts.size() ; ++i)
    {
      if (this->_counts[i] == 0.0 )
	res.push_back(this->_fracExp[i]);
      else
	res.push_back( this->_counts[i] / this->_fracExp[i]);
    }

  _fracExpWeight = res;
}


std::vector<double> LCGrid::getFracExpWeight()
{
	return _fracExpWeight;
}



void LCGrid::setCCDNumber(int ccd)
{
	_ccd = ccd;
}

int LCGrid::getCCDNumber()
{
	return _ccd;
}

bool LCGrid::checkBinning(LCGrid* bkgLC)
{
	bool check = true;
	errstr << "Checking Source and Background Binning"
		<<Sas::message(Msg::App,Msg::Sparse);
	errstr << "Source Start Time: "<< setprecision(16) << this->_start
		<<Sas::message(Msg::App,Msg::Sparse);
	errstr << "Background Start Time: "<< setprecision(16) << bkgLC->_start
			<<Sas::message(Msg::App,Msg::Sparse);
	if (abs(this->_start-bkgLC->_start) >1E-5)
	 {
		 check = false;
		 errstr << "Time Serie Start Time not aligned."
		 	<< Sas::message(Msg::App,Msg::Sparse);
	 }
	errstr <<"Source Stop Time: "<< setprecision(16) << this->_stop
		<<Sas::message(Msg::App,Msg::Sparse);
	errstr <<"Background Stop Time: "<< setprecision(16) << bkgLC->_stop
			<<Sas::message(Msg::App,Msg::Sparse);

	if (abs(this->_stop-bkgLC->_stop) > 1E-5)
	 {
		 check = false;
		 errstr << "Time Serie Stop Time not aligned."
		 	<< Sas::message(Msg::App,Msg::Sparse);
	 }

	return check;
}

std::vector<double> operator*(std::vector<double> v1,double n)
{
	std::vector<double> res;
	for (unsigned int i = 0 ; i < v1.size() ; i++)
		res.push_back(v1[i]*n);

	return res;
}

std::vector<double> operator-(std::vector<double> v1,std::vector<double> v2)
{
	std::vector<double> res;
	for (unsigned int i = 0 ; i < v1.size() ; i++)
		res.push_back(v1[i]-v2[i]);

	return res;
}

std::vector<double> operator+(std::vector<double> v1,std::vector<double> v2)
{
	std::vector<double> res;
	for (unsigned int i = 0 ; i < v1.size() ; i++)
		res.push_back(v1[i]+v2[i]);

	return res;
}

std::vector<double> operator*(std::vector<double> v1,std::vector<double> v2)
{
	std::vector<double> res;
	for (unsigned int i = 0 ; i < v1.size() ; i++)
		res.push_back(v1[i]*v2[i]);

	return res;
}

std::vector<double> operator/(std::vector<double> v1,std::vector<double> v2)
{
	std::vector<double> res;
	res.clear();
	for (unsigned int i = 0 ; i < v1.size() ; i++)
	{
	  double val = v2[i] == 0.0 ? 0.0 : v1[i] / v2[i];
	  res.push_back(val);
	}

	return res;
}

std::vector<double> operator/(std::vector<double> v1,double v2)
{
	std::vector<double> res;
	res.clear();
	for (unsigned int i = 0 ; i < v1.size() ; i++)
	{
	  double val = v2 == 0.0 ? 0.0 : v1[i] / v2;
	  res.push_back(val);
	}

	return res;
}
