// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.


#include <iomanip>
#include <iostream>
#include <DataFilter.h>
#include <evselect.h>
#include <ParameterInfo.h>
#include <ExceptionErrorHandler.h>
#include <cmath>
#include "Xmm.h"
#include "NAN.h"
#include <cmath>

#include <unistd.h>
//gcc4.8.2 portability changes
#include <cstdio>

#include "stringutil.h"
#include "LightCurve.h"
#include "FrameTable.h"
#include "GtiCorr.h"
#include "Exposure.h"
#include "DataSubspace.h"

LightCurve::LightCurve(string timeSerie, string evtList)
{
	_bkgCorrected = false;

	_LC = setLC(timeSerie);
	_LC->_name = timeSerie;
	setDSS();
	closeTimeSerie();
	setEventList(evtList);
	setCCDsInRegion();
	closeEventList();
	if (_ccdInRegion.size() >1 )
	{
		// The original Time Serie has not got any CCD attached.
		// So, we set the 0 to this particular Time Serie.
		_LC->setCCDNumber(0);
		// We only recalculate the CCDs time series for those cases where
		// the selection region overlap more than one CCD
		createCCDsLC();
	}
	else
		_LC->setCCDNumber(_ccdInRegion[0]);

}

LightCurve::LightCurve(string timeSerie)
{
	setLC(timeSerie);
	setDSS();
	closeTimeSerie();
	_evtExists = false;
	_bkgCorrected = false;

}

LightCurve::LightCurve(double val0, double val1, double binSize)
{
	errstr << "creating a simple light curve" << Sas::message(Msg::App,Msg::Sparse);
	_LC = new LCGrid(val0,val1,binSize);
	_tsExists = false;
	_evtExists = false;
	_bkgCorrected = false;

}

LightCurve::~LightCurve()
{
	errstr << "LightCurve Destructor "<< Sas::message(Msg::App,Msg::Sparse);
	delete _LC;

	if (_bkgCorrected)
	  delete _spec;

	for (unsigned int i = 0 ; i < _spectVector.size() ; ++i)
	  delete _spectVector[i];

	for (unsigned int i = 0 ; i < _ccdLC.size() ; i++)
	{
	  dataSetServer->remove(_ccdLC[i]->_name);
	  delete _ccdLC[i];
	}


	errstr << "Done "<< Sas::message(Msg::App,Msg::Sparse);
}


LCGrid* LightCurve::setLC(string timeSerie)
{
	_tsExists = true;
	// Read the Time serie file
	_ts =  dataSetServer->open(timeSerie, Dal::Read);

	// Read from the Time Serie header the First bin, the lastest bin and
	// the bin size.
	// We cannot use the TSTART and the TSTOP keywords, because they are
	// the time stamp of the first and the latest event or the timemin and timemax values of
	// evselect parameters.
	// TIMEDEL == Light curve bin size


	_obs_id = _ts->attribute("OBS_ID")->asString();
	_exp_id = _ts->attribute("EXP_ID")->asString();
	_observer = _ts->attribute("OBSERVER")->asString();
	_object = _ts->attribute("OBJECT")->asString();
	_exp_id_str= _ts->attribute("EXPIDSTR")->asString();

	Table *tsTab = _ts->table("RATE");

	// Set the INSTRUMENT and TIME keywords attribute
	_instrument = tsTab->attribute("INSTRUME")->asString();
	_filter = tsTab->attribute("FILTER")->asString();
	_telescope = tsTab->attribute("TELESCOP")->asString();
	_mjdref = tsTab->attribute("MJDREF")->asReal();
	_timeSys = tsTab->attribute("TIMESYS")->asString();
	_timeSysComment = tsTab->attribute("TIMESYS")->label();
	_timeUnit = tsTab->attribute("TIMEUNIT")->asString();
	_timeRef = tsTab->attribute("TIMEREF")->asString();
	_timeRefComment = tsTab->attribute("TIMEREF")->label();
	_timeAssign = tsTab->attribute("TASSIGN")->asString();
	_timeAssignComment = tsTab->attribute("TASSIGN")->label();
	_exposure = tsTab->attribute("EXPOSURE")->asReal();
	_timeClockApp = tsTab->attribute("CLOCKAPP")->asString();
	_tstart = tsTab->attribute("TSTART")->asReal();
	_tstop = tsTab->attribute("TSTOP")->asReal();

	ColumnData *timeDat = tsTab->column("TIME")->data();
	real64 *time = timeDat->real64Data();


	double binSize = tsTab->attribute("TIMEDEL")->asReal();
	// LCGrid needs as input the start time of the time serie, the stop time and the bin size.
	// Remember that the TIME column is the time bin middle value.
	size_t nrows = tsTab->rows();
	LCGrid *lc = new LCGrid(time[0]-(binSize/2.0),time[tsTab->rows()-1]+(binSize/2.0),nrows);

	// Store the RATE or the COUNT values

	string colName = tsTab->hasColumn("RATE") ? "RATE" : "COUNTS";
	ColumnData *rateDat = tsTab->column(colName)->data();
	if ( colName == "RATE" )
	{
		real64 *rate = rateDat->real64Data();
		lc->setCounts(rate);
	}
	else // Assume COUNTS column
	{
		int32 *counts = rateDat->int32Data();

		lc->setCounts(counts);
	}

	// Setting the DataSubspace;

	delete rateDat;
	delete timeDat;	

	return lc;

}

void LightCurve::setDSS()
{
	Block * eblk = _ts->block(_ts->name()+":RATE");
	eblk >> _dss;

}

void LightCurve::setEventList(string evtList)
{
	_evtExists = true;
	_evtListName = evtList;
	dataSetServer->copy(evtList, "tmp.ds");
	_evtList = dataSetServer->open("tmp.ds", Dal::Modify);
}

void LightCurve::setCCDsInRegion()
{

	Table *srcTable = _evtList->table("EVENTS");

	//Source filtering

	DataFilter filter(_dss.selectExpression());
	unsigned long selectedEvents = filter.doFilter(*srcTable);

	_subMode = srcTable->attribute("SUBMODE")->asString();

	errstr << "selected " << selectedEvents << " rows from the source input table."
	  	     << Sas::message(Msg::App,Msg::Sparse);

	ColumnData * cCcdDat = srcTable->column("CCDNR")->data();
	int8 * ccdDat = cCcdDat->int8Data();

	unsigned long mRows = srcTable->rows();

	int ccdNum=0;
	for(unsigned long irow=0; irow<mRows; irow++){
		if (ccdNum != ccdDat[irow])
			_ccdInRegion.push_back(ccdDat[irow]);

		ccdNum = ccdDat[irow];
	}

	delete cCcdDat;

}

void LightCurve::createCCDsLC()
{
  /*
   * Create the light curve per CCD and the spectrum per CCD
   */
	for (unsigned int i = 0 ; i < _ccdInRegion.size() ; i++)
	{
		// Create a lightcurve for only one CCD
		// Check if there is any region in the DSS
		std::stringstream o;

		if (_dss.selectExpression().size() ==  0)
		{
			errstr << "No selection region found. Using all CCD." <<
					Sas::message(Msg::App,Msg::Sparse);
		}
		else
		{
			o << "( "<<_dss.selectExpression() << ") ";
					o << "&& ";
		}

		o << "(CCDNR == "<< _ccdInRegion[i] << ") ";
		string expression = o.str();
		Info::Task* evselect = task("evselect");
		ParameterInfo& evselectpars = *evselect->parameters();
		evselectpars.revertToDefault();

		evselectpars["table"]->setValue(_evtListName+ ":EVENTS");
		evselectpars["expression"]->setValue(expression);
		evselectpars["writedss"]->setValue(true);
		evselectpars["updateexposure"]->setValue(false);
		evselectpars["withrateset"]->setValue(true);
		std::stringstream ccd;

		ccd << tmpnam(0)<<"rate_ccd" << _ccdInRegion[i]<<".fits";
		evselectpars["rateset"]->setValue(ccd.str());
		evselectpars["timebinsize"]->setValue(_LC->_stride);
		evselectpars["withtimeranges"]->setValue(true);

		real64 timemin = _LC->_start ;
		real64 timemax = _LC->_stop;
		evselectpars["timemin"]->setValue(timemin);
		evselectpars["timemax"]->setValue(timemax );
		evselectpars["maketimecolumn"]->setValue(true);
		evselectpars["makeratecolumn"]->setValue(true);

		/*
		 * Spectrum
		 */
		std::stringstream specName;
		specName << tmpnam(0)<< "spectrum_ccd" << _ccdInRegion[i]<<".fits";
		evselectpars["energycolumn"]->setValue("PI");
		evselectpars["withspectrumset"]->setValue(true);
		evselectpars["spectrumset"]->setValue(specName.str());
		evselectpars["spectralbinsize"]->setValue(50);
		try{
		    evselect->exec(true);
		    _ccdLC.push_back(setLC(ccd.str()));
		    // Set the CCD where this LC belongs;
		    _ccdLC[i]->setCCDNumber(_ccdInRegion[i]);
		    _ccdLC[i]->_name = ccd.str();
		    // Create spectrum.
		    Spect *spec = new Spect(specName.str());
		    spec->setEventFileName(_evtListName);
		    _spectVector.push_back(spec);
		  }
		  catch (SasException::Error e) {
		    errstr << evselect->name() << " terminated with the error " << e.code
			   << " that said: \n" << e.message
			   << Sas::warning("SubTaskError");

		  }
		  catch (SasException::Fatal e) {
		    errstr << evselect->name() << " terminated with the fatal error " << e.code
			   << " that said: \n" << e.message
			   << Sas::warning("SubTaskError");
		  }
	}
}

void LightCurve::doRelativeCorr(Spect::arfgenType val)
{
	//Check if we have more than one CCD in the selection region.

	if (_ccdInRegion.size() > 1)
	{
		errstr << "More than one CCD inside the extraction region" <<
		Sas::message(Msg::App,Msg::Sparse);
		// Check that we have the sum of the CCDs LC Counts is the same as the
		// Counts in the original CCD

		std::vector<double> ccdCountVal =  _LC->getCounts();
		for (unsigned int i = 0 ; i < _ccdInRegion.size() ; i++)
		  ccdCountVal = ccdCountVal - _ccdLC[i]->getCounts();


		for (unsigned int i = 0 ; i < ccdCountVal.size() ; i++)
		{
		  if ( ccdCountVal[i] >= 0.5 )
		    {
		      errstr << "A difference in the sum of the counts of single CCDs " <<
			"with respect to the counts of the original light curve has been found.  "
			     << Sas::warning("CCDs Count Difference");
		    }
		}

		// Apply the relative correction for each CCDs Light Curves
		errstr << "Calculating the relative corrections for each CCD..." <<
		  Sas::message(Msg::App,Msg::Sparse);

		for (unsigned int i = 0 ; i < _ccdInRegion.size() ; i++)
		{
		  errstr << "Correction for CCD: "<< _ccdLC[i]->_ccd <<
		    Sas::message(Msg::App,Msg::Sparse);
		  doGTICorr(_ccdLC[i]);

		  //If there isn't any selection reigion, we assume
		  //that the user wants only relative corrections.

		  if (booleanParameter("allcamera") ==  false)
			  doCCDCorrection(_spectVector[i],val);

		  // double area=_spectVector[i]->getArea();


		}


		errstr << "Done." << Sas::message(Msg::App,Msg::Sparse);


		//if there isn't any extraction region, we do not have to weight the areas

		if (booleanParameter("allcamera") ==  false)
		{

			// Calculate the weigth that we will use afterwards to calculate the
			// FracExp of the corresponding original Light Curve.

			// Add the corresponding FracExp Weight of each CCD Light Curve
			errstr << "Calculating the final fracExp value..." <<
					Sas::message(Msg::App,Msg::Sparse);

			/*
			 * if the user select withbkgsubtraction, we have to apply the effective area
			 * correction. if the user only wants to correct a source light curve, we only correct
			 * the different CCD light curve weighting with each corresponging BACKSCAL
			 */
			// Apply effective area correction factor for all CCDs
			std::vector<double> fracExp;
			double areaTot;

			if (val == Spect::SRC )
			{
				errstr << "using effective area..." <<
						Sas::message(Msg::App,Msg::Sparse);
				double area=_spectVector[0]->getArea();
				doExposureCorr(_ccdLC[0],area);
				fracExp = _ccdLC[0]->_fracExp;
				areaTot = _spectVector[0]->getArea();
				for (unsigned int i = 1 ; i < _ccdLC.size() ; ++i)
				{
					area=_spectVector[i]->getArea();
					doExposureCorr(_ccdLC[i],area);
					fracExp = fracExp +_ccdLC[i]->_fracExp;
					areaTot = areaTot +_spectVector[i]->getArea();
				}
				if (areaTot == 0.0)
					errstr << "Sum of all CCD areas is empty. Please check your source and background expression"
					<< Sas::error("ZeroAreaCorrection");

				fracExp = fracExp / areaTot;
			}
			else
			{
				errstr << "using geometrical area..." <<
						Sas::message(Msg::App,Msg::Sparse);
				double area=_spectVector[0]->getBackscale();
				doExposureCorr(_ccdLC[0],area);
				fracExp = _ccdLC[0]->_fracExp;
				areaTot = _spectVector[0]->getBackscale();
				for (unsigned int i = 1 ; i < _ccdLC.size() ; ++i)
				{
					area=_spectVector[i]->getBackscale();
					doExposureCorr(_ccdLC[i],area);
					fracExp = fracExp +_ccdLC[i]->_fracExp;
					areaTot = areaTot +_spectVector[i]->getBackscale();
				}

				if (areaTot == 0.0)
					errstr << "Sum of all CCD areas is empty. Please check your source and background expression"
					<< Sas::error("ZeroAreaCorrection");

				fracExp = fracExp / areaTot;

			}

			errstr << "Done." << Sas::message(Msg::App,Msg::Sparse);
			// Set the new FracExp to the original LC
			_LC->setFracExp(fracExp);
		}
		else
		{
			std::vector<double> fracExp;
			doExposureCorr(_ccdLC[0],1.0);
			double areaTot = 1.0;
			fracExp = _ccdLC[0]->_fracExp;
			for (unsigned int i = 1 ; i < _ccdLC.size() ; ++i)
			{
				doExposureCorr(_ccdLC[i],1.0);
				fracExp = fracExp +_ccdLC[i]->_fracExp;
				areaTot = areaTot + 1.0;
			}
			fracExp = fracExp / areaTot;

			errstr << "Done." << Sas::message(Msg::App,Msg::Sparse);
			// Set the new FracExp to the original LC
			_LC->setFracExp(fracExp);
		}

	}
	else
	{
		errstr << "Extraction region lies in CCD number " << _LC->getCCDNumber()
					<< Sas::message(Msg::App,Msg::Sparse);
		errstr << "Calculating the relative corrections for each CCD..." <<
				Sas::message(Msg::App,Msg::Sparse);

		doGTICorr(_LC);
		doExposureCorr(_LC,1.0);
		errstr << "Relative Corrections Done." << Sas::message(Msg::App,Msg::Sparse);

	}

}


void LightCurve::mergeFrames()
{
  FrameList res;
  FrameList ffl = _expVector[0].getFrameList();
  FrameList ifl = _expVector[1].getFrameList();
  OutOGIPgtiTable out1("gti1.ds");
  out1<<ffl._frameList;

  OutOGIPgtiTable out2("gti2.ds");
  out2<<ifl._frameList;

   ffl._frameList |= ifl._frameList;
   ffl._frameList.sortIt();
   res = ffl;
   Exposure exposure(ffl);

   _expVector.push_back(exposure);

  OutOGIPgtiTable out3("gti3.ds");
  out3<<res._frameList;



}

void LightCurve::doExpCorrMultpleCCDs(LCGrid *lc)
{

  GtiCorr gti(_expVector[0].getFrameList()._frameList);
  gti.doCorr(lc);
  //_expVector[2].doCorr(lc);
}


void LightCurve::doGTICorr(LCGrid *lc)
{
	std::vector<std::string> nameExtVec = searchFor(lc);

	// Set up GTI info
	// Combine all the GTI extension of the Time Serie
	errstr << "Doing GTI Correction..."<< Sas::message(Msg::App,Msg::Sparse);
	// Loop over all the GTI extension and combine all into a single
	// OGIPgtiTable.
	GtiList li1;

	for (unsigned int i = 0 ; i < nameExtVec.size() ; i++)
	{
		errstr << "GTI Extension["<<nameExtVec[i] <<"]"
			<< Sas::message(Msg::App,Msg::Sparse);

		InOGIPgtiTable in1 = InOGIPgtiTable(lc->_name+":"+nameExtVec[i]);
		GtiList li; in1>>li;
		if (i == 0)
			li1 = li;
		else
			li1 |= li;
	}

	li1.sortIt();

	GtiCorr gti(li1);
	gti.doCorr(lc);
	gti.printCorr();

	errstr << "Done." << Sas::message(Msg::App,Msg::Sparse);

}


void LightCurve::doExposureCorr(LCGrid *lc,double area)
{
	std::stringstream ccd;
	bool randFlag;
	ccd << ":EXPOSU";
	lc->getCCDNumber() < 10 ? ccd <<"0"<<lc->getCCDNumber() : ccd << lc->getCCDNumber();

	errstr  << "Doing Exposure and Dead Time Correction: "
	<< _evtListName+ccd.str() <<  Sas::message(Msg::App,Msg::Sparse);

	FrameTable in2(_evtListName+ccd.str());
	FrameList li2; in2>>li2;

	/*
	 * Check if the binsize is smaller than frametime
	 */
 	if (lc->_stride < in2.getFrameTime() )
		errstr << "Light curve bin size smaller than frame time."
		<< Sas::error("BinSizeTooSmall");


	/*
	 * Check if the events have been randomize
	 */
	randFlag = in2.getRandFlag();
	if (randFlag)
	{
		errstr  << "Events already randomized..."<<  Sas::message(Msg::App,Msg::Sparse);
	}


	if ( (_instrument == "EPN") && (randFlag == false) )
	{
		if ( (Xmm::toMode(_subMode) ==  Xmm::FAST_TIMING) || (Xmm::toMode(_subMode) ==  Xmm::FAST_BURST )  )
		{
			errstr << "Randomization for PN Timing or Burst not implemented"
			<< Sas::warning("PNTimingRand");
		}
	}

	in2.closeDataSet();

	Exposure exposure(li2);
	//_expVector.push_back(exposure);

	exposure.doCorr(lc,area);
	//exposure.printCorr();

	errstr << "Exposure Correction Done." << Sas::message(Msg::App,Msg::Sparse);


	/*
	 * Dead Time Correction
	 */
	/* Dead Time correction removed because within the exposure correction we are already
	 * correcting for the dead time. Nevertheless, we will need this correction to create
	 * new columns giving information about the deadtime fracexp values.
	 */
	//errstr  << "Doing Dead Time Correction "<<  Sas::message(Msg::App,Msg::Sparse);
	//DeadTime deadTime(li2);
	//deadTime.doCorr(lc);
	//deadTime.printCorr();
	//errstr << "Dead Time Correction Done." << Sas::message(Msg::App,Msg::Sparse);
	/*
	 * Dead Time Correction
	 */

}


void LightCurve::substractBKGForAllCCDs(LightCurve *bkg)
{

  errstr  << "Substracting background for each CCDs..."<<  Sas::message(Msg::App,Msg::Sparse);

  std::vector<double> bkgRate = bkg->_LC->_rateCorr;
  double timeBinSize = bkg->_LC->_stride;
  if ( (this->_spec->getBackscale() == 0) || ( bkg->_spec->getBackscale() == 0) )
    errstr  << "Zero Source or Background backscale. "<<  Sas::error("ZeroBackscale");

  double backScaleRat = (double)this->_spec->getBackscale()/(double)bkg->_spec->getBackscale();
  this->_spec->setBackscaleRatio(backScaleRat);

  std::vector<std::vector<double> > ccdsErr;
  for (unsigned int i = 0 ; i < _ccdLC.size() ; ++i)
  {
    double backScale =  (double)_spectVector[i]->getBackscale() / (double)bkg->_spec->getBackscale();

    std::vector<double> srcCts = _ccdLC[i]->getCounts();
    std::vector<double> errVal;

    for (unsigned int j = 0 ; j < srcCts.size() ; ++j)
    {
      _ccdLC[i]->_counts[j] = srcCts[j] - (bkgRate[j]*backScale*timeBinSize*_ccdLC[i]->_fracExpNoArea[j]);
      double e = sqrt(srcCts[j] + (backScale*backScale)* bkgRate[j]*timeBinSize*_ccdLC[i]->_fracExpNoArea[j]);

      //cout << "CCD ERR VALS "<< srcCts[j]<< " bkgRate "<< bkgRate[j]  << " backscale "<< backScale << " FracExp "<< _ccdLC[i]->_fracExpNoArea[j]<< " eror " << e << endl;
      errVal.push_back(e);
    }
    ccdsErr.push_back(errVal);
  }
  // reset the background subtracted counts
  std::vector<double> srcBkgSubCts =_ccdLC[0]->getCounts();
  std::vector<double> errTotal = ccdsErr[0];
  for (unsigned int i = 0 ; i < _ccdLC.size() ; ++i)
    {
      std::vector<double> errTotal = ccdsErr[i];
      //for (unsigned int k = 0 ; k <= errTotal.size() ; ++k)
	//cout << "LL "<< i << "  " << errTotal[k] << endl;
    }

  errTotal = (ccdsErr[0]*ccdsErr[0]) + (ccdsErr[1]*ccdsErr[1]) + (ccdsErr[2]*ccdsErr[2]);
  for (unsigned int i = 1 ; i < _ccdLC.size() ; ++i)
  {
	  srcBkgSubCts = srcBkgSubCts + _ccdLC[i]->getCounts();
	  //errTotal = (errTotal * errTotal) + (ccdsErr[i]*ccdsErr[i]);
  }
  _LC->_counts =  srcBkgSubCts;


  //rate and Errors
	_LC->_rateCorr.clear();
	_LC->_errRateCorr.clear();
	_LC->_timeStamp.clear();
	double rate, error;

	for (unsigned int i = 0 ; i <= this->_LC->_cells ; i++)
	{
		rate = this->_LC->_fracExp[i] == 0.0 ? 0.0 : this->_LC->_counts[i]/(this->_LC->_stride *this->_LC->_fracExp[i]);
		error = this->_LC->_fracExp[i] == 0.0 ? 0.0 : sqrt(errTotal[i])/(this->_LC->_stride *this->_LC->_fracExp[i]);
		//cout << "ERROR "<< errTotal[i] << " " << this->_LC->_stride << " " << this->_LC->_fracExp[i] << " "  <<error << endl;
		_LC->_rateCorr.push_back(rate);
		_LC->_errRateCorr.push_back(error);
		_LC->_timeStamp.push_back((*this->_LC)[i]);
	}
	//rate and errors



    errstr  << "Done."<<  Sas::message(Msg::App,Msg::Sparse);
}


std::vector<std::string> LightCurve::searchFor(LCGrid *lc)
{
	std::vector<std::string> values;
	DataSet *ts  =  dataSetServer->open(lc->_name, Dal::Read);
	DataSubspace dss;
	Block * eblk = ts->block(ts->name()+":RATE");
	eblk >> dss;


	unsigned ncomps = dss.components();

	for(unsigned icomp=0; icomp<ncomps; icomp++)
	{
		DScomp * comp = dss.component(icomp);
		for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++){
			DSfilter * filter = comp->filter(ifilt);

			if( (filter->type()==DSS::GtiT)	 )
    	    {
    	    	std::vector<std::string> vec(breakup(filter->blockSpec(), ":"));
    	    	// Check if we have a STDGTI block or a GTI block

    	    	std::stringstream ss;
    	    	int number;
    	    	if (vec[1].find("STDGTI") != string::npos)
    	    	{
    	    		errstr  << vec[1] <<" extension found "<<  Sas::message(Msg::App,Msg::Sparse);
        	    	ss << vec[1][vec[1].size()-2] << vec[1][vec[1].size()-1];
        	    	ss >> number;
    	    	}
    	    	else
    	    	{
    	    		Block* expBlock = ts->block(vec[1]);
    	    		errstr  << vec[1] <<" extension found "<<  Sas::message(Msg::App,Msg::Sparse);
    	    		if (!expBlock->hasAttribute("CCDID"))
    	    		{
    	    			errstr  << "No CCDID keyword found in "<<
							vec[1]<< " table. Please, recreate the eventlist, "<<
							" srctslist and bkgtslist files with latest evselect."<<
							Sas::error("NoCCDID");
    	    		}
    	    		else
    	    			number = expBlock->attribute("CCDID")->asInt();


    	    		//ss << vec[1][vec[1].size()-5] << vec[1][vec[1].size()-4];
    	    		//ss >> number;
        	    	// I do not know why, with GTI extensions, the ccd number is decreased by 1.
        	    	// New dsslib version 5.6 -> fixes this problem that has to do with the
        	    	// GTI naming. Now, all the GTI extensions has its own CCD number in the table
        	    	// name, then it is simply to identify the GTI that you have to use.
        	    	// AI: Fixes SOC-SPR-6503
        	    	//number++;
    	    		expBlock->release();
    	    	}
                if (number == lc->getCCDNumber())
                {
                	errstr  <<"Processing GTI extension: "<< vec[1] <<
                	" corresponding to CCD " << number <<
                	Sas::message(Msg::App,Msg::Sparse);
					values.push_back(vec[1]);
                }

    	    }
		}
	}

	dataSetServer->close(ts);

	if (values.size() == 0)
		errstr << "No GTI found for the CCD number "<< lc->getCCDNumber()
			<< Sas::error("EmptyGTIList");

	return values;
}

void LightCurve::closeTimeSerie()
{
	if (_tsExists == true )
		dataSetServer->close(_ts);
}

void LightCurve::closeEventList()
{
	errstr << "Closing event file dataset"<< Sas::message(Msg::App,Msg::Sparse);
	if (_evtExists == true)
	{
		dataSetServer->close(_evtList);
		unlink("tmp.ds");
	}
}
LCGrid* LightCurve::getLC()
{
	return _LC;
}

void LightCurve::calSpec()
{
	errstr << "Calculating spectrum..."<< Sas::message(Msg::App,Msg::Sparse);
	_bkgCorrected = true;

	// Creates an spectrum. Store all the information to be used for the
	// absolute correction.
	Info::Task* evselect = task("evselect");
	ParameterInfo& evselectpars = *evselect->parameters();
	evselectpars.revertToDefault();

	evselectpars["table"]->setValue(_evtListName+ ":EVENTS");
	evselectpars["writedss"]->setValue(true);
	std::vector<std::string> vec = breakup(_LC->_name, "/");
	std::string name = vec[vec.size()-1];

	std::stringstream specName;
	specName << "spectrum_" << name;
	evselectpars["expression"]->setValue(_dss.selectExpression());
	evselectpars["energycolumn"]->setValue("PI");
	evselectpars["withspectrumset"]->setValue(true);
	evselectpars["spectrumset"]->setValue(specName.str());
	evselectpars["spectralbinsize"]->setValue(50);
	try{
	    evselect->exec(true);
	    // Create spectrum.
	    _spec = new Spect(specName.str());
	    _spec->setEventFileName(_evtListName);
	}
	catch (SasException::Error e) {
		errstr << evselect->name() << " terminated with the error " << e.code
		   << " that said: \n" << e.message
		   << Sas::warning("SubTaskError");

	}
	catch (SasException::Fatal e) {
		errstr << evselect->name() << " terminated with the fatal error " << e.code
		   << " that said: \n" << e.message
		   << Sas::warning("SubTaskError");
	}

	errstr << "Done."<< Sas::message(Msg::App,Msg::Sparse);
}

Spect* LightCurve::getSpec()
{
	return _spec;
}


void LightCurve::doAbsoluteCorr()
{
	/*
	 * Calculate the average of the off/on axis ARF, weighted with the spectrum
	 * counts.
	 */
	std::vector<double> ratio = _spec->_onSpecResp/_spec->_srcSpecResp;

	/*
	 * Convert the Channel into energy (keV)
	 */
	double energy;

	std::vector<double> absCorr;
	std::vector<double> weights;
	for (unsigned int i = 0 ; i < _spec->_channels.size() ; i++)
	{
		unsigned int energyCounter = 0;
		energy = ((i*_spec->_specdelt)+_spec->_specval)/1000.0;

		while ((!_spec->_energyList[energyCounter].inside(energy))
				&& (energyCounter < ratio.size()))
			energyCounter++;

		if (energyCounter <  ratio.size())
		{
			absCorr.push_back(_spec->_counts[i]*ratio[energyCounter+1]);
			weights.push_back(_spec->_counts[i]);
		}

	}


	/*
	 * Weight the ratio with the spectrum counts
	 */
	double val = 0.0;
	double w = 0.0;
	for (unsigned int i = 0 ; i < absCorr.size() ; i++)
	{
		val += absCorr[i];
		w += weights[i];
	}
	_spec->setWeigthedMean( val/w);

	errstr << "Weighted Mean "<< _spec->getWeigthedMean() << Sas::message(Msg::App,Msg::Sparse);
	if ( _spec->getWeigthedMean() == 0.0)
		errstr << "Zero Absolute Correction. Please check your source and background expression"
			<< Sas::error("ZeroAbsoluteCorrection");
}

void LightCurve::doCCDCorrection(Spect *sp,Spect::arfgenType val)
{
  errstr  <<"Doing CCD corrections... " << Sas::message(Msg::App,Msg::Sparse);
  /*
   * Run arfgen for each ccd to calculate the effective area c
   *
   */

  sp->calArfgen(val);
  sp->readSpec();
  sp->readArf(val);
  if (val == Spect::BKG)
	  sp->removeBkgArfFile();
  else
	  sp->removeArfFile();
  /*
   * Convert the Channel into energy (keV)
   */
	double energy;

	std::vector<double> absCorr;
	std::vector<double> weights;
	for (unsigned int i = 0 ; i < sp->_channels.size() ; i++)
	{
		unsigned int energyCounter = 0;
		energy = ((i*sp->_specdelt)+sp->_specval)/1000.0;

		while ((!sp->_energyList[energyCounter].inside(energy))
				&& (energyCounter < sp->_srcSpecResp.size()))
			energyCounter++;

		if (energyCounter <  sp->_srcSpecResp.size())
		{
			absCorr.push_back(sp->_counts[i]*sp->_srcSpecResp[energyCounter+1]);
			weights.push_back(sp->_counts[i]);
		}

	}


	/*
	 * Weight the ratio with the spectrum counts
	 */
	double value = 0.0;
	double w = 0.0;
	for (unsigned int i = 0 ; i < absCorr.size() ; i++)
	{
		value += absCorr[i];
		w += weights[i];
	}
	if (value <= 0)
	{
		errstr << "Problems gathering information from arfgen "<<
				" Setting coeffs values to 0" << Sas::warning("wrongArfgenResults");
		value= 0.0;
	}
	sp->setArea( value/w);

	errstr << "Area in CCD "<< sp->getArea() << Sas::message(Msg::App,Msg::Sparse);
	if ( sp->getArea() == 0.0)
		errstr << "Zero Effective area. Please check your source and background expression"
			<< Sas::warning("ZeroArfgenArea");



}


void LightCurve::applyAbsCorr()
{
	errstr << "Applying Absolute Correction... "   << Sas::message(Msg::App,Msg::Sparse);
	//_LC->_counts = _LC->_counts * _spec->getWeigthedMean();

	_LC->_rateCorr = _LC->_rateCorr * _spec->getWeigthedMean();
	_LC->_errRateCorr = _LC->_errRateCorr * _spec->getWeigthedMean();
	errstr << "Done. "   << Sas::message(Msg::App,Msg::Sparse);
}

bool LightCurve::moreThanOneCCD()
{
  if (_ccdInRegion.size() > 1)
    return true;
  else
    return false;
}


void LightCurve::createGti(LightCurve *lc, string name, DataSet *set)
{
	errstr << "Creating " << name << " GTIs..."
			<< Sas::message(Msg::App,Msg::Sparse);
	//GTI extension that contains the NULL values
	GtiList src_gti_list;
	std::vector<unsigned int> startRows;
	std::vector<unsigned int> finisRows;
	bool withinGti = false;

	for (unsigned int i = 0 ; i < lc->_LC->_timeStamp.size();i++)
	{
		if ( lc->_LC->_fracExp[i] == 0.0)
		{
			if (withinGti)
			{
				finisRows.push_back(i-1);
				withinGti = false;
			}
		}
		else
		{
			if (!withinGti)
			{
				startRows.push_back(i);
				withinGti = true;
			}
		}
	}
	if (withinGti)
	{
		finisRows.push_back(lc->_LC->_timeStamp.size()-1);
	}

	for (unsigned int i = 0 ; i < startRows.size();i++)
	{
		double start=lc->_LC->_timeStamp[startRows[i]] - (lc->_LC->_stride / 2.0);
		double stop=lc->_LC->_timeStamp[finisRows[i]] + (lc->_LC->_stride / 2.0);
		Gti src_gti_vals( start,stop);
		src_gti_list.push_back(src_gti_vals);
	}

	Table *srcGtiTab = set->addTable(name,startRows.size());
	OutOGIPgtiTable tempgti(srcGtiTab);

	tempgti << src_gti_list;

	errstr << "Done." << Sas::message(Msg::App,Msg::Sparse);


}

void LightCurve::createFile(string outFileName, LightCurve *bkgLC)
{
	DataSet *out  =  dataSetServer->open(outFileName, Dal::Create);
	Table *outTable = out->addTable("RATE",_LC->_timeStamp.size());

	// Create columns
	real64 *timeDat=0;
	real32 *rateDat=0, *rateErrDat=0, *fracExpDat=0;
	ColumnData *cTimeDat=0, *cRateDat=0, *cRateErrDat=0, *cFracExpDat=0;
	real32 *bkgRateDat=0, *bkgRateErrDat=0;
	ColumnData *cBkgRateDat=0 , *cBkgRateErrDat=0;
	cTimeDat = outTable->addColumn("TIME", Dal::Real64, "Time of the center of each frame", "s")->data();
	cRateDat = outTable->addColumn("RATE", Dal::Real32, "Source Count Rate", "count/s")->data();
	cRateErrDat = outTable->addColumn("ERROR", Dal::Real32, "Error in Source light curve", "count/s")->data();
	cFracExpDat = outTable->addColumn("FRACEXP", Dal::Real32, "effective time / real time", "fraction")->data();

	timeDat = cTimeDat->real64Data();
	rateDat = cRateDat->real32Data();
	rateErrDat = cRateErrDat->real32Data();
	fracExpDat = cFracExpDat->real32Data();


	/*
	 * if Background defined
	 */
	if (bkgLC != 0)
	{
		cBkgRateDat = outTable->addColumn("BACKV", Dal::Real32, "Bkg rate normalized to src area", "count/s")->data();
		cBkgRateErrDat = outTable->addColumn("BACKE", Dal::Real32, "Error in bkg normalized to src area", "count/s")->data();
		bkgRateDat = cBkgRateDat->real32Data();
		bkgRateErrDat = cBkgRateErrDat->real32Data();
	}

	for (unsigned int i = 0 ; i < _LC->_timeStamp.size();i++)
	{
		timeDat[i] = _LC->_timeStamp[i];
		fracExpDat[i] = _LC->_fracExp[i];
		if ( _LC->_fracExp[i] == 0.0)
		{
			NaN.set(rateDat[i]);
			NaN.set(rateErrDat[i]);

			if (bkgLC != 0)
			{
				NaN.set(bkgRateDat[i]);
				NaN.set(bkgRateErrDat[i]);
			}

		}
		else {
			rateDat[i] = _LC->_rateCorr[i];
			rateErrDat[i] = _LC->_errRateCorr[i];
			if (bkgLC != 0)
			{
				//Normalized Background rate and error
				bkgRateDat[i] = bkgLC->getLC()->_rateCorr[i]*this->_spec->getBackscaleRatio();
				bkgRateErrDat[i] = bkgLC->getLC()->_errRateCorr[i]*this->_spec->getBackscaleRatio();
			}
		}
	}

	createGti(this,"SRC_GTIS",out);

	if (bkgLC != 0)
		createGti(bkgLC,"BKG_GTIS",out);

	// Attributes
	*out->addAttribute("TELESCOP","XMM Mission" )=_telescope;
	*out->addAttribute( "INSTRUME",
				"Instrument name")= _instrument;
	*out->addAttribute("OBS_ID","Observation Identifier" )=_obs_id;
	*outTable->addAttribute("EXP_ID","Exposure Identifier" )=_exp_id;
	*out->addAttribute("OBJECT","Name of observed object" )=_object;
	*out->addAttribute("OBSERVER","Name of observer" )=_observer;
	*out->addAttribute("EXPIDSTR","Exposure ID" )=_exp_id_str;

	*outTable->addAttribute("TELESCOP","XMM Mission" )=_telescope;
	*outTable->addAttribute( "INSTRUME",
				"Instrument name")= _instrument;
	*outTable->addAttribute("OBS_ID","Observation Identifier" )=_obs_id;
	*outTable->addAttribute("EXP_ID","Exposure Identifier" )=_exp_id;
	*outTable->addAttribute("OBJECT","Name of observed object" )=_object;
	*outTable->addAttribute("OBSERVER","Name of observer" )=_observer;
	*outTable->addAttribute("EXPIDSTR","Exposure ID" )=_exp_id_str;
	*outTable->addAttribute("FILTER","Filter ID" )=_filter;
	*outTable->addAttribute("MJDREF","1998-01-01T00:00:00 (TT) expressed in MJD (TT)","d" )= _mjdref;
	*outTable->addAttribute( "HDUCLASS",
				"Format conforms to OGIP/GSFC conventions")= "OGIP";

	*outTable->addAttribute( "HDUCLAS1",
				"File contains a time-serie")= "LIGHTCURVE";

	*outTable->addAttribute("TSTART","Light curve start time" )=_tstart;
	*outTable->addAttribute("TSTOP","Light curve end time" )=_tstop;

	std::string key, message;
	if (booleanParameter("withbkgset"))
	{
		key = "NET";
		message = "File contains bkg subtracted rates";
	}
	else
	{
		key = "TOTAL";
		message = "File contains src+bkg rates";
	}

	*outTable->addAttribute( "HDUCLAS2",
				message)= key;

	*outTable->addAttribute( "HDUCLAS3",
				"Light curve stored as rates")= "RATE";
	*outTable->addAttribute( "HDUVERS1",
				"Version of format")= "1.1.0";


	//
	// Write DSS to fits file
	//
	*outTable->addAttribute("SLCTEXPR", "Filtering expression used by "
					     "evselect") = _dss.selectExpression();


	//
	// TIME Attributes
	//
	*outTable->addAttribute( "TIMESYS",
					_timeSysComment)= _timeSys;
	*outTable->addAttribute( "TIMEUNIT",
						"All times in s unless specified otherwise")= _timeUnit;
	*outTable->addAttribute( "TIMEREF",
						_timeRefComment)= _timeRef;
	*outTable->addAttribute( "TASSIGN",
						_timeAssignComment)= _timeAssign;
	*outTable->addAttribute( "CLOCKAPP",
						"Clock correction applied")= _timeClockApp;

	//Delete GTI blocks
	errstr << "Removing GTI information from DSS..."   << Sas::message(Msg::App,Msg::Sparse);
	unsigned ncomps = _dss.components();

	for(unsigned icomp=0; icomp<ncomps; icomp++){
		DScomp * comp = _dss.component(icomp);

		if( comp->hasFilter("TIME", DSS::RangeT) || comp->hasFilter("TIME", DSS::RangeBlockT) ||
				comp->hasFilter("TIME", DSS::GtiT) ){
			//_dss.deleteComponent(icomp);
			for (unsigned ifilt=0; ifilt<comp->filters() ; ifilt++)
			{
				DSfilter *filter = comp->filter(ifilt);
				if(filter->name()=="TIME"){
					if(filter->type()==DSS::GtiT){
						comp->deleteFilter(ifilt);
					}
				}
			}
			ncomps = _dss.components();
			icomp = 0;
	      }

	    }
	outTable << _dss;
	errstr << "Done"   << Sas::message(Msg::App,Msg::Sparse);

	//Extract the CHANMIN and CHANMAX values from the DSS
	errstr << "Extracting PI information from DSS..."   << Sas::message(Msg::App,Msg::Sparse);
	double lower=0.0;
	double upper=0.0;
	bool chanflag = false;
	for(unsigned icomp=0; icomp<ncomps; icomp++){
		DScomp * comp = _dss.component(icomp);
		if( comp->hasFilter("PI", DSS::RangeT) )
			for (unsigned ifilt=0; ifilt<comp->filters() ; ifilt++)
			{
				DSfilter *filter = comp->filter(ifilt);
				if(filter->name()=="PI"){
					if(filter->type()==DSS::RangeT){
						std::list<Range<double> > ranges =filter->ranges();
						std::list<Range<double> >::const_iterator irange;
						for(irange=ranges.begin(); irange!=ranges.end(); irange++){
							RangeBound lowerB, upperB;
							irange->values(lower, upper, lowerB, upperB);
							if( lowerB==Unbounded) lower = 0.0;
							if( upperB==Unbounded ) upper = -1;
							chanflag=true;
						}
					}
				}
			}
	}
	errstr << "Done"   << Sas::message(Msg::App,Msg::Sparse);
	//Get the PI min and max values


	// Rate file specific keywords.

	*outTable->addAttribute( "TIMEDEL",
				"Light curve bin size")= _LC->_stride;
	*outTable->addAttribute( "TSTART",
				"Light curve start time")= _LC->_start;

	*outTable->addAttribute( "TSTOP",
				"Light curve end time")= _LC->_stop;
	*outTable->addAttribute( "BACKAPP", "Background correction applied?") = booleanParameter("withbkgset");
	*outTable->addAttribute( "DEADAPP", "Dead time correction applied?") = true;
	*outTable->addAttribute( "VIGNAPP", "Vignetting correction applied?") = booleanParameter("applyabsolutecorrections");

	if (booleanParameter("withbkgset"))
		*outTable->addAttribute( "BKGRATIO", "Ratio of BKG to SRC collection areas") = 1. / this->_spec->getBackscaleRatio();

	if (chanflag)
	{
		*outTable->addAttribute( "CHANMIN",
				"Minimum value of the energy column")= lower;
		if (upper != -1)
		{
			*outTable->addAttribute( "CHANMAX",
					"Maximum value of the energy column")= upper;
		}
		*outTable->addAttribute( "CHANTYPE",
				"Name of energy column used")= "PI";
	}
	else
		errstr << "No PI information in the DSS"
		<< Sas::warning("NoPIInfoInDss");


	*outTable->addAttribute("EXPOSURE","Weighted live time of CCDs in the extraction re") = _exposure;

	delete cTimeDat;
	delete cRateDat;
	delete cRateErrDat;
	delete cBkgRateDat;
	delete cBkgRateErrDat;
	delete cFracExpDat;

	dataSetServer->close(out);

}


LightCurve* LightCurve::operator-(LightCurve* bkg)
{
	errstr << "Substracting background... "   << Sas::message(Msg::App,Msg::Sparse);
	//Check that the Lightcurve binning are aligned
	if (!this->getLC()->checkBinning(bkg->getLC()))
		errstr  << "Source and Background Lightcurve not aligned. Please, re-calculate "
				"the time series using the same start and stop time"<<  Sas::error("Binning");

	// Check that the backscale has been calculated and get the keywords.

	errstr << "Source BACKSCALE: "<< this->_spec->getBackscale() << Sas::message(Msg::App,Msg::Sparse);
	errstr << "Background BACKSCALE: "<< bkg->_spec->getBackscale() << Sas::message(Msg::App,Msg::Sparse);
	if ( (this->_spec->getBackscale() == 0) || ( bkg->_spec->getBackscale() == 0) )
		errstr  << "Zero Source or Background backscale. "<<  Sas::error("ZeroBackscale");

	double backScale = (double)this->_spec->getBackscale() / (double)bkg->_spec->getBackscale();

	this->_spec->setBackscaleRatio(backScale);

	this->getLC()->_rateCorr = this->getLC()->_rateCorr - ( bkg->getLC()->_rateCorr * backScale );


	for (unsigned int i = 0 ; i < this->getLC()->_errRateCorr.size() ; i++)
	{
		this->getLC()->_errRateCorr[i] =
			sqrt( (double)  std::abs( (this->getLC()->_errRateCorr[i] * this->getLC()->_errRateCorr[i]) +
					( ( bkg->getLC()->_errRateCorr[i] * backScale ) * ( bkg->getLC()->_errRateCorr[i] * backScale ) ) ) );
	}




	errstr << "Done. "   << Sas::message(Msg::App,Msg::Sparse);
	return this;

}

