// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

#include <unistd.h>
#include <Dal.h>
#include <arfgen.h>
#include <ExceptionErrorHandler.h>
#include "errstr.h"
#include "Spect.h"


Spect::Spect(std::string name) :
	_specName(name)
{

}

Spect::~Spect()
{
	unlink(_specName.c_str());
}

void Spect::calArfgen(arfgenType mode)
{
	errstr << "Calculating arfgen..."<< Sas::message(Msg::App,Msg::Sparse);

	// Creates an spectrum. Store all the information to be used for the
	// absolute correction.
	Info::Task* arfgen = task("arfgen");
	ParameterInfo& arfgenpars = *arfgen->parameters();
	arfgenpars.revertToDefault();
	switch ( mode )
	{
	case SRC:
		arfgenpars["spectrumset"]->setValue(_specName);
		arfgenpars["arfset"]->setValue("src_ancillary.ds");
		arfgenpars["setbackscale"]->setValue(true);
		arfgenpars["badpixlocation"]->setValue(_eventFileName);

		if(booleanParameter("withdetbins"))
		{
			arfgenpars["withdetbins"]->setValue(true);
			arfgenpars["detxbins"]->setValue(intParameter("detxbins"));
			arfgenpars["detybins"]->setValue(intParameter("detybins"));
		}

		if(booleanParameter("withsourcepos"))
		{
			arfgenpars["withsourcepos"]->setValue(true);
			arfgenpars["sourcecoords"]->setValue(stringParameter("sourcecoords"));
			arfgenpars["sourcex"]->setValue(realParameter("sourcex"));
			arfgenpars["sourcey"]->setValue(realParameter("sourcey"));
		}
		break;
	case ON:
		arfgenpars["arfset"]->setValue("on_ancillary.ds");
		arfgenpars["spectrumset"]->setValue(_specName);
		arfgenpars["withsourcepos"]->setValue(true);
		arfgenpars["sourcecoords"]->setValue("tel");
		arfgenpars["sourcex"]->setValue(0.0);
		arfgenpars["sourcey"]->setValue(0.0);
		arfgenpars["filterdss"]->setValue(false);
		arfgenpars["withdetbounds"]->setValue(true);
		arfgenpars["withbadpixcorr"]->setValue(false);
		arfgenpars["modelee"]->setValue(false);
		arfgenpars["detmaptype"]->setValue("flat");
		arfgenpars["detxbins"]->setValue(1);
		arfgenpars["detybins"]->setValue(1);
		break;
	case BKG:
		arfgenpars["spectrumset"]->setValue(_specName);
		arfgenpars["arfset"]->setValue("bkg_ancillary.ds");
		arfgenpars["setbackscale"]->setValue(true);
		arfgenpars["badpixlocation"]->setValue(_eventFileName);
		arfgenpars["extendedsource"]->setValue(true);
		break;
	}


	try{
	    arfgen->exec(true);
	    // The bkg ancillary file is not needed. We only have to run arfgen with bkg file to
	    // get the BACKSCALE.

	}
	catch (SasException::Error e) {
		errstr << arfgen->name() << " terminated with the error " << e.code
		   << " that said: \n" << e.message
		   << Sas::warning("SubTaskError");

	}
	catch (SasException::Fatal e) {
		errstr << arfgen->name() << " terminated with the fatal error " << e.code
		   << " that said: \n" << e.message
		   << Sas::warning("SubTaskError");
	}
	errstr << "Done."<< Sas::message(Msg::App,Msg::Sparse);
}

void Spect::readSpec()
{
	errstr << "Reading Spectrum... "<< _specName
		<< Sas::message(Msg::App,Msg::Sparse);
	DataSet *specSet  =  dataSetServer->open(_specName, Dal::Read);
	Table *specTab = specSet->table("SPECTRUM");
	setBackscale(specTab->attribute("BACKSCALE")->asInt());
	_specval = specTab->attribute("SPECVAL")->asReal();
	_specpix = specTab->attribute("SPECPIX")->asInt();
	_specdelt = specTab->attribute("SPECDELT")->asInt();

	ColumnData *channelsDat = specTab->column("CHANNEL")->data();
	int16 *channels = channelsDat->int16Data();

	ColumnData *countsDat = specTab->column("COUNTS")->data();
	int32 *counts = countsDat->int32Data();

	for (unsigned int i = 0 ; i < specTab->rows() ; i++)
	{
		_channels.push_back(channels[i]);
		_counts.push_back(counts[i]);
	}

	delete channelsDat;
	delete countsDat;

	dataSetServer->close(specSet);


}

void Spect::readArf(arfgenType mode)
{
	errstr << "Reading ancillary file... "<< _specName
		<< Sas::message(Msg::App,Msg::Sparse);
	DataSet *set = 0;
	Table *arfTab;
	ColumnData *specRespDat = 0;
	ColumnData *eLowDat = 0;
	ColumnData *eHighDat = 0;
	real32 *specResp;
	real32 *eLow, *eHigh;
	switch ( mode )
	{
		case SRC :
			set =  dataSetServer->open("src_ancillary.ds", Dal::Read);
			arfTab = set->table("SPECRESP");
			specRespDat = arfTab->column("SPECRESP")->data();
			eLowDat =  arfTab->column("ENERG_LO")->data();
			eHighDat =  arfTab->column("ENERG_HI")->data();
			specResp = specRespDat->real32Data();
			eLow = eLowDat->real32Data();
			eHigh = eHighDat->real32Data();
			for (unsigned int i = 0 ; i < arfTab->rows() ; i++)
			{
				_srcSpecResp.push_back(specResp[i]);
				Gti gti(eLow[i],eHigh[i]);
				_energyList.push_back(gti);
			}

			break;
		case ON:
			set  =  dataSetServer->open("on_ancillary.ds", Dal::Read);
			arfTab = set->table("SPECRESP");
			specRespDat = arfTab->column("SPECRESP")->data();
			eLowDat =  arfTab->column("ENERG_LO")->data();
			eHighDat =  arfTab->column("ENERG_HI")->data();
			specResp = specRespDat->real32Data();
			eLow = eLowDat->real32Data();
			eHigh = eHighDat->real32Data();

			for (unsigned int i = 0 ; i < arfTab->rows() ; i++)
			{
				_onSpecResp.push_back(specResp[i]);
				Gti gti(eLow[i],eHigh[i]);
				_energyList.push_back(gti);
			}

			unlink("on_ancillary.ds");

			break;
		case BKG:
			set =  dataSetServer->open("bkg_ancillary.ds", Dal::Read);
			arfTab = set->table("SPECRESP");
			specRespDat = arfTab->column("SPECRESP")->data();
			eLowDat =  arfTab->column("ENERG_LO")->data();
			eHighDat =  arfTab->column("ENERG_HI")->data();
			specResp = specRespDat->real32Data();
			eLow = eLowDat->real32Data();
			eHigh = eHighDat->real32Data();
			for (unsigned int i = 0 ; i < arfTab->rows() ; i++)
			{
				_srcSpecResp.push_back(specResp[i]);
				Gti gti(eLow[i],eHigh[i]);
				_energyList.push_back(gti);
			}

//			errstr <<"BKG must not read the ancillary file. "<< Sas::error("BKG arfgen error");
			break;
	}
	delete specRespDat;
	delete eLowDat;
	delete eHighDat;
	dataSetServer->close(set);

	errstr << "Done."<< Sas::message(Msg::App,Msg::Sparse);
}

// We need to remove the file in case the user
// does not want to apply the abs correction.
// We need to run arfgen to get the right
// backscale.
void Spect::removeArfFile()
{
	unlink("src_ancillary.ds");
}

void Spect::removeBkgArfFile()
{
	unlink("bkg_ancillary.ds");
}

void Spect::setEventFileName(std::string evtName)
{
	_eventFileName = evtName;
}

void Spect::setBackscale(int val)
{
	_backscale = val;
}

void Spect::setBackscaleRatio(double val)
{
	_backscaleRatio = val;
}

double Spect::getBackscaleRatio()
{
	return _backscaleRatio;
}

int Spect::getBackscale()
{
	return _backscale;
}

void Spect::setWeigthedMean(double val)
{
	_weigthedMean = val;
}

double Spect::getWeigthedMean()
{
	return _weigthedMean;
}

void Spect::setArea(double area)
{
  _area = area;
}

double Spect::getArea()
{
  return _area;
}

